﻿local addonName, addon = ...

local tsort = table.sort

local LibQTip = LibStub('LibQTip-1.0')

local playerKey = UnitName('player') .. "@" .. GetRealmName()
local selectedKey = playerKey

local characterClass, characterLists, instanceData, realmList, selectedData
local indexLUT, resetLUT = { }

local GetResetLFR = addon.GetResetLFR
local resetLFR = GetResetLFR()
addon.GetResetLFR = nil

--[[----------------------------------------------------------------------------
Cooldown data management
------------------------------------------------------------------------------]]
local sortAscend, sortPriority = { true, true, true }, { 1, 3, 0, 2 }
local numSortPriority, dataKey = #sortPriority
sortAscend[0] = true

local function AlphaSort(a, b)
	return a:lower() < b:lower()
end

local function CooldownSort(a, b)
	local field, A, B
	for index = 1, numSortPriority do
		field = sortPriority[index]
		if field > 1 then
			A, B = a[field], b[field]
		elseif field == 1 then
			A, B = a[field]:lower(), b[field]:lower()
		else
			A, B = resetLUT[a], resetLUT[b]
		end
		if A ~= B then
			if sortAscend[field] then
				return A < B
			else
				return A > B
			end
		end
	end
end

local function UpdateData()
	local dataExpired

	local time = time()
	for reset in pairs(instanceData) do
		if reset <= time then
			dataExpired, instanceData[reset] = true, nil
		end
	end

	if dataExpired or not realmList then
		local playerClass = characterClass[playerKey]
		characterClass[playerKey] = nil
		for key, class in pairs(characterClass) do
			local hasData
			for _, resetData in pairs(instanceData) do
				if resetData[key] then
					hasData = true
					break
				end
			end
			if not hasData then
				characterClass[key] = nil
			end
		end
		characterClass[playerKey] = playerClass
		if not characterClass[selectedKey] then
			selectedKey = playerKey
		end

		characterLists, realmList = { }, { }
		local list
		for key, class in pairs(characterClass) do
			local name, realm = key:match("^(.+)@(.+)$")
			list = characterLists[realm]
			if not list then
				list = { }
				characterLists[realm], realmList[#realmList + 1] = list, realm
			end
			list[#list + 1] = name
		end
		tsort(realmList, AlphaSort)
		for _, list in pairs(characterLists) do
			tsort(list, AlphaSort)
		end
	end

	if dataKey ~= selectedKey or dataExpired then
		local count = 0
		resetLUT, selectedData = { }, { }
		for reset, resetData in pairs(instanceData) do
			local data = resetData[selectedKey]
			if data then
				for index = 1, #data do
					count = count + 1
					resetLUT[data[index]], selectedData[count] = reset, data[index]
				end
			end
		end
		if count > 1 then
			tsort(selectedData, CooldownSort)
		end
		dataKey = selectedKey
	end

	return dataExpired
end

local function AddInfoToResetData(reset, ...)
	local data, info = instanceData[reset], { ... }
	if not data then
		data = { }
		instanceData[reset] = data
	end
	data = data[playerKey]
	if not data then
		data = { }
		instanceData[reset][playerKey] = data
	end
	data[#data + 1] = info
	return info
end

--[[----------------------------------------------------------------------------
settings.instanceData[resetTime][playerKey] = {
	{ name, ID, difficultyCode, difficultyName }, -- [1]
}
------------------------------------------------------------------------------]]
local function SaveInstanceInfo()
	for _, resetData in pairs(instanceData) do
		resetData[playerKey] = nil
	end
	wipe(indexLUT)

	local time = time() + 30
	for index = 1, GetNumSavedInstances() do
		local name, lockID, reset, difficulty, locked, _, _, _, maxPlayers, difficultyName = GetSavedInstanceInfo(index)
		if locked and reset > 5 then
			local reset = floor((time + reset) / 60) * 60
			indexLUT[AddInfoToResetData(reset, name, format("%08X", lockID), maxPlayers * 1000 + difficulty, difficultyName)] = index
		end
	end

	for index = 1, GetNumSavedWorldBosses() do
		local name, _, reset = GetSavedWorldBossInfo(index)
		if reset > 5 then
			local reset = floor((time + reset) / 60) * 60
			AddInfoToResetData(reset, name, "", 0, RAID_INFO_WORLD_BOSS)
		end
	end

	if resetLFR < time then
		resetLFR = GetResetLFR()
	end
	for index = 1, GetNumRFDungeons() do
		local dungeonID = GetRFDungeonInfo(index)
		local numEncounters, numCompleted = GetLFGDungeonNumEncounters(dungeonID)
		if numCompleted > 0 then
			local name, _, _, _, _, _, _, _, _, _, _, difficulty, maxPlayers = GetLFGDungeonInfo(dungeonID)
			local info = AddInfoToResetData(resetLFR, name, "", maxPlayers * 1000 + difficulty, RAID_FINDER_PVEFRAME)
			for bossID = 1, numEncounters do
				local bossName, _, isKilled = GetLFGDungeonEncounterInfo(dungeonID, bossID)
				if isKilled then
					info[#info + 1] = bossName
				end
			end
		end
	end

	for reset, resetData in pairs(instanceData) do
		if not next(resetData) then
			instanceData[reset] = nil
		end
	end

	if selectedKey == playerKey then
		dataKey = nil
		UpdateData()
	end
end

--[[----------------------------------------------------------------------------
Character select menu
------------------------------------------------------------------------------]]
local function Checked(self)
	if self.arg2 then
		self.arg2 = nil
		return self.arg1 == selectedKey
	end
end

local function OnClick(self, key)
	if key ~= selectedKey then
		local name, realm = key:match("^(.+)@(.+)$")
		local oldName, oldRealm = selectedKey:match("^(.+)@(.+)$")
		if realm == oldRealm then
			local list = characterLists[realm]
			for index = 1, #list do
				if list[index] == oldName then
					local dropDownButton = ("DropDownList%dButton%d"):format(DropDownList2:IsShown() and 2 or 1, index)
					_G[dropDownButton .. 'Check']:Hide()
					_G[dropDownButton .. 'UnCheck']:Show()
					break
				end
			end
		end
		selectedKey = key
		UpdateData()
	end
end

local menu = LibStub('LibMenuAssist-1.0'):New()
menu.xOffset, menu.yOffset = 0, 0

menu.initialize = function(self, level)
	local info, UIDropDownMenu_AddButton = UIDropDownMenu_CreateInfo(), UIDropDownMenu_AddButton
	if level == 1 and #realmList > 1 then
		info.hasArrow, info.notCheckable = true, true
		for index = 1, #realmList do
			info.text = realmList[index]
			UIDropDownMenu_AddButton(info, level)
		end
	else
		info.checked, info.func, info.keepShownOnClick, info.arg2 = Checked, OnClick, true, true
		local format, realm = string.format, level > 1 and UIDROPDOWNMENU_MENU_VALUE or realmList[1]
		local colors, list, color, key = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS, characterLists[realm]
		for index = 1, #list do
			key = list[index] .. "@" .. realm
			color = colors[characterClass[key]]
			info.text = format("|cff%02x%02x%02x%s", color.r * 255, color.g * 255, color.b * 255, list[index])
			info.arg1 = key
			UIDropDownMenu_AddButton(info, level)
		end
	end
end

--[[----------------------------------------------------------------------------
Tooltip stuff
------------------------------------------------------------------------------]]
local tooltip, LoadTooltip

local function RepositionAlternateTooltip()
	GameTooltip:ClearAllPoints()
	if tooltip:GetCenter() >= UIParent:GetCenter() then
		GameTooltip:SetPoint('RIGHT', tooltip, 'LEFT', 0, 0)
	else
		GameTooltip:SetPoint('LEFT', tooltip, 'RIGHT', 0, 0)
	end
end

local function RaidFinder_OnEnter(self, data)
	GameTooltip:SetOwner(tooltip, 'ANCHOR_NONE')
	RepositionAlternateTooltip()
	GameTooltip:AddLine(ERR_LOOT_GONE)
	for index = 5, #data do
		GameTooltip:AddLine(data[index], 1, 1, 1)
	end
	GameTooltip:Show()
end

local function SavedInstance_OnEnter(self, data)
	GameTooltip:SetOwner(tooltip, 'ANCHOR_NONE')
	GameTooltip:SetInstanceLockEncountersComplete(indexLUT[data])
	if GameTooltipTextLeft2:IsShown() then
		RepositionAlternateTooltip()
		GameTooltipTextLeft1:SetText(data[1])
	else
		GameTooltip:Hide()
		indexLUT[data] = nil
		LoadTooltip()
	end
end

local function OnLeave(self)
	GameTooltip:Hide()
end

local function OnMouseUp(self, field)
	if not self:IsMouseOver() then return end
	if sortPriority[1] == field then
		sortAscend[field] = not sortAscend[field]
	else
		sortAscend[field] = true
		for index = 2, numSortPriority do
			if sortPriority[index] == field then
				tremove(sortPriority, index)
				tinsert(sortPriority, 1, field)
				break
			end
		end
	end
	if not UpdateData() and #selectedData > 1 then
		tsort(selectedData, CooldownSort)
	end
	LoadTooltip()
end

function LoadTooltip()
	tooltip:Clear()
	tooltip:SetColumnLayout(4, 'CENTER', 'CENTER', 'CENTER', 'CENTER')

	local color = (CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS)[characterClass[selectedKey]]
	tooltip:SetCell(tooltip:AddHeader(" "), 1, ("|cff%02x%02x%02x%s |cff808080(%s)"):format(color.r * 255, color.g * 255, color.b * 255, selectedKey:match("^(.+)@(.+)$")), 'CENTER', 4)

	tooltip:AddLine(" ")
	tooltip:AddHeader("|cffffd200Instance/Boss", "|cffffd200Type", "|cffffd200Cooldown", "|cffffd200ID")

	local lineNum = tooltip:GetLineCount()
	tooltip:SetCellScript(lineNum, 1, 'OnMouseUp', OnMouseUp, 1)
	tooltip:SetCellScript(lineNum, 2, 'OnMouseUp', OnMouseUp, 3)
	tooltip:SetCellScript(lineNum, 3, 'OnMouseUp', OnMouseUp, 0)
	tooltip:SetCellScript(lineNum, 4, 'OnMouseUp', OnMouseUp, 2)

	tooltip:AddSeparator(2)
	tooltip:SetColumnLayout(4, 'LEFT', 'LEFT', 'CENTER', 'LEFT')

	local time = time()
	for index = 1, #selectedData do
		local data = selectedData[index]
		tooltip:AddLine(data[1], data[4], SecondsToTime(resetLUT[data] - time), data[2])
		local OnEnter = (data[5] and RaidFinder_OnEnter) or (indexLUT[data] and SavedInstance_OnEnter)
		if OnEnter then
			local lineNum = tooltip:GetLineCount()
			tooltip:SetLineScript(lineNum, 'OnEnter', OnEnter, data)
			tooltip:SetLineScript(lineNum, 'OnLeave', OnLeave)
		end
	end
	if #selectedData == 0 then
		tooltip:SetCell(tooltip:AddLine(" "), 1, "<empty>", 'CENTER', 4)
	end

	tooltip:UpdateScrolling()
	tooltip:SetClampedToScreen(true)
end

local function OnRelease()
	tooltip = nil
end

--[[----------------------------------------------------------------------------
LDB Object
------------------------------------------------------------------------------]]
local function OnEnter(self)
	menu:Close()
	if not tooltip then
		tooltip = LibQTip:Acquire(addonName)
		tooltip:EnableMouse(true)
		if self.AttachFrame then
			self:AttachFrame(tooltip)
		else
			tooltip:SmartAnchorTo(self)
		end
		tooltip.OnRelease = OnRelease
		tooltip:SetAutoHideDelay(1, self)
	end
	if selectedKey == playerKey then
		RequestRaidInfo()
	end
	UpdateData()
	LoadTooltip()
	tooltip:Show()
end

LibStub('LibDataBroker-1.1'):NewDataObject(addonName, {
	type = 'launcher',
	icon = [[Interface\Icons\Spell_Holy_BorrowedTime]],
	label = addonName,

	OnClick = function(self, button)
		if button == 'LeftButton' then
			if selectedKey ~= playerKey then
				selectedKey = playerKey
				OnEnter(self)
			end
		elseif button == 'RightButton' then
			if tooltip then
				tooltip:Hide()
			end
			local _, y = self:GetCenter()
			if y > UIParent:GetHeight() / 2 then
				menu.point, menu.relativeTo, menu.relativePoint = 'TOP', self, 'BOTTOM'
			else
				menu.point, menu.relativeTo, menu.relativePoint = 'BOTTOM', self, 'TOP'
			end
			UpdateData()
			menu:Open()
			menu:AddFocusFrame(self)
		end
	end,

	OnEnter = OnEnter,

	OnLeave = function(self, motion)
		if menu:IsShown() then
			menu:RemoveAllFocusFrames()
		end
		if tooltip and not motion then
			tooltip:Release()
		end
	end
})

--[[----------------------------------------------------------------------------
Initialize / Throttle data saves
------------------------------------------------------------------------------]]
local frame = CreateFrame('Frame')
frame:Hide()

frame:SetScript('OnEvent', function(self, event, name)
	if name ~= addonName then return end
	self:UnregisterEvent('ADDON_LOADED')

	local settings = _G[addonName .. "Settings"]
	if type(settings) ~= 'table' then
		settings = { }
		_G[addonName .. "Settings"] = settings
	end
	settings.version = GetAddOnMetadata(addonName, 'Version')

	instanceData = settings.instanceData
	if type(instanceData) ~= 'table' then
		instanceData = { }
		settings.instanceData = instanceData
	end

	characterClass = settings.characterClass
	if type(characterClass) ~= 'table' then
		characterClass = { }
		settings.characterClass = characterClass
	end
	characterClass[playerKey] = select(2, UnitClass('player'))

--	Throttle cooldown data saves
	self:SetScript('OnEvent', function(self, event)
		if event == 'PLAYER_ENTERING_WORLD' or event == 'PLAYER_LOGOUT' then
			RequestRaidInfo()
		else
			self:Show()
		end
	end)
	self:RegisterEvent('PLAYER_ENTERING_WORLD')
	self:RegisterEvent('PLAYER_LOGOUT')
	self:RegisterEvent('UPDATE_INSTANCE_INFO')

	self:SetScript('OnUpdate', function(self, elapsed)
		self:Hide()
		SaveInstanceInfo()
		if tooltip then
			LoadTooltip()
		end
	end)

	hooksecurefunc('SetSavedInstanceExtend', function()
		RequestRaidInfo()
	end)
end)
frame:RegisterEvent('ADDON_LOADED')
